from collections.abc import ItemsView, Iterator, KeysView, ValuesView

import pydantic


class ListOf[T](pydantic.RootModel[list[T]]):
    """
    Pydantic RootModel for representing a list of other models.
    """

    def __iter__(self) -> Iterator[T]:  # type:ignore[override]
        return iter(self.root)

    def __getitem__(self, item: int) -> T:
        return self.root[item]

    def __len__(self) -> int:
        return len(self.root)


class DictOf[K, V](pydantic.RootModel[dict[K, V]]):
    """
    Pydantic RootModel for representing a dict
    """

    def __getitem__(self, key: K) -> V:
        return self.root[key]

    def __setitem__(self, key: K, value: V) -> None:
        self.root[key] = value

    def __delitem__(self, key: K) -> None:
        del self.root[key]

    def __contains__(self, key: K) -> bool:
        return key in self.root

    def __iter__(self) -> Iterator[K]:  # type:ignore[override]
        return iter(self.root)

    def __len__(self) -> int:
        return len(self.root)

    def keys(self) -> KeysView[K]:
        return self.root.keys()

    def values(self) -> ValuesView[V]:
        return self.root.values()

    def items(self) -> ItemsView[K, V]:
        return self.root.items()

    def get(self, key: K, default: V | None = None) -> V | None:
        return self.root.get(key, default)

    def setdefault(self, key: K, default: V) -> V:
        return self.root.setdefault(key, default)

    def update(self, other: dict[K, V]) -> None:
        self.root.update(other)

    def pop(self, key: K, default: V | None = None) -> V | None:
        return self.root.pop(key, default)

    def popitem(self) -> tuple[K, V]:
        return self.root.popitem()

    def clear(self) -> None:
        self.root.clear()



#!/usr/bin/env python
import os
import sys


def main():
    os.environ.setdefault(
        "DJANGO_SETTINGS_MODULE", "thelabtyping.tests.settings.sqlite"
    )

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)


if __name__ == "__main__":
    main()



from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.test import Client, TestCase
from django.urls import reverse

from thelabtyping.api.routing import HttpMethod, Router
from thelabtyping.api.status import Status

from ..sampleapp.views import (
    route_users,
    router,
    router_users_list,
)


class RouterTest(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(username="joe", password="password")
        self.client = Client()

    def test_reverse_router_urls(self) -> None:
        self.assertEqual(
            reverse("sampleapp:router-users-list"),
            "/api/router/users/",
        )
        self.assertEqual(
            reverse("sampleapp:router-users-detail", args=(42,)),
            "/api/router/users/42/",
        )
        self.assertEqual(
            reverse("sampleapp:alt-router-users-list"),
            "/api/alt-router/users/",
        )
        self.assertEqual(
            reverse("sampleapp:alt-router-users-detail", args=(42,)),
            "/api/alt-router/users/42/",
        )

    def test_router_index_main(self) -> None:
        url = reverse("sampleapp:index")
        self.assertEqual(url, "/api/router/")
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            {
                "sampleapp:index": "http://testserver/api/router/",
                "sampleapp:router-users-list": "http://testserver/api/router/users/",
            },
        )

    def test_router_index_unnamespaced(self) -> None:
        url = reverse("unnamespaced-router-index")
        self.assertEqual(url, "/unnamespaced-router/")
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            {
                "unnamespaced-router-index": "http://testserver/unnamespaced-router/",
                "unnamespaced-router-users-list": "http://testserver/unnamespaced-router/users/",
            },
        )

    def test_router_dispatch_get(self) -> None:
        resp = self.client.get(reverse("sampleapp:router-users-list"))
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            [
                {
                    "username": "joe",
                    "first_name": "",
                    "last_name": "",
                }
            ],
        )

    def test_router_dispatch_post(self) -> None:
        self.client.login(username="joe", password="password")
        resp = self.client.post(
            reverse("sampleapp:router-users-list"),
            content_type="application/json",
            data={
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            {
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )

    def test_router_dispatch_method_not_allowed(self) -> None:
        self.client.login(username="joe", password="password")
        resp = self.client.put(reverse("sampleapp:router-users-list"))
        self.assertEqual(resp.status_code, Status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_register_duplicate_route(self) -> None:
        with self.assertRaises(ImproperlyConfigured):
            router.route("users/", name="router-users-list-duplicate")

    def test_register_duplicate_route_method(self) -> None:
        with self.assertRaises(ImproperlyConfigured):
            route_users.get(router_users_list)

    def test_register_get(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.get(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.GET})

    def test_register_post(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.post(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.POST})

    def test_register_put(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.put(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.PUT})

    def test_register_patch(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.patch(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.PATCH})

    def test_register_delete(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.delete(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.DELETE})

    def test_register_during_route_creation(self) -> None:
        router = Router()
        route = router.route(
            "test/",
            name="test",
            get=router_users_list,
            post=router_users_list,
        )
        self.assertEqual(
            route.allowed_methods,
            {
                HttpMethod.GET,
                HttpMethod.POST,
            },
        )



from collections.abc import Callable
from dataclasses import dataclass
from enum import StrEnum
from functools import reduce
from operator import or_
from typing import Concatenate, NewType

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponseBase, HttpResponseNotAllowed
from django.urls import URLPattern, path, reverse
from django.urls.exceptions import NoReverseMatch
import pydantic

from ..abc import DictOf
from .responses import APIResponse


class HttpMethod(StrEnum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


URLPatternStr = NewType("URLPatternStr", str)
AllowedMethods = set[HttpMethod]

type ViewFn[**P, R: HttpResponseBase] = Callable[Concatenate[HttpRequest, P], R]

ALL_METHODS: AllowedMethods = {
    HttpMethod.GET,
    HttpMethod.POST,
    HttpMethod.PUT,
    HttpMethod.PATCH,
    HttpMethod.DELETE,
}


@dataclass
class RegisteredView[**P, R: HttpResponseBase]:
    methods: AllowedMethods
    fn: ViewFn[P, R]


@dataclass
class Route:
    name: str
    views: list[RegisteredView[[], HttpResponseBase]]

    def __init__(self, name: str) -> None:
        self.name = name
        self.views = []

    def get[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        return self.register({HttpMethod.GET})(fn)

    def post[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        return self.register({HttpMethod.POST})(fn)

    def put[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        return self.register({HttpMethod.PUT})(fn)

    def patch[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        return self.register({HttpMethod.PATCH})(fn)

    def delete[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        return self.register({HttpMethod.DELETE})(fn)

    def register[**P, R: HttpResponseBase](
        self,
        methods: AllowedMethods = ALL_METHODS,
    ) -> Callable[
        [ViewFn[P, R]],
        ViewFn[P, R],
    ]:
        def decorator(view_fn: ViewFn[P, R]) -> ViewFn[P, R]:
            self.add_view(view_fn, methods)
            return view_fn

        return decorator

    def add_view[**P, R: HttpResponseBase](
        self,
        view_fn: ViewFn[P, R],
        methods: AllowedMethods,
    ) -> None:
        conflicts = self.allowed_methods & methods
        if conflicts:
            raise ImproperlyConfigured(
                f"Cannot {view_fn} for methods {conflicts}. Views already "
                "exist for these methods."
            )
        self.views.append(
            RegisteredView(
                methods=methods,
                fn=view_fn,
            )
        )

    def dispatch(
        self,
        request: HttpRequest,
        *args: object,
        **kwargs: object,
    ) -> HttpResponseBase:
        for view in self.views:
            if request.method in view.methods:
                return view.fn(request, *args, **kwargs)

        return HttpResponseNotAllowed(self.allowed_methods)

    @property
    def allowed_methods(self) -> AllowedMethods:
        return reduce(or_, (view.methods for view in self.views), set())


type RouteMap = dict[URLPatternStr, Route]


RouterIndex = DictOf[str, pydantic.HttpUrl]


class Router:
    basename: str | None = None
    routes: RouteMap

    def __init__(
        self,
        basename: str | None = None,
        enable_index: bool = True,
    ) -> None:
        self.basename = basename
        self.routes: RouteMap = {}
        # Immediately register the root index view
        if enable_index:
            self.route("", name="index", get=self.index_view)

    def index_view(self, request: HttpRequest) -> APIResponse[RouterIndex]:
        index = RouterIndex({})
        namespace = request.resolver_match.namespace if request.resolver_match else None
        for pattern, route in self.routes.items():
            name = f"{namespace}:{route.name}" if namespace else route.name
            try:
                url_path = reverse(name)
            except NoReverseMatch:
                # Catch and ignore this so that we skip URLs which require
                # params (e.g. detail views)
                continue
            url = request.build_absolute_uri(url_path)
            index[name] = pydantic.HttpUrl(url)
        return APIResponse(index)

    def route[**P, R: HttpResponseBase](
        self,
        url_pattern: str,
        name: str,
        get: ViewFn[P, R] | None = None,
        post: ViewFn[P, R] | None = None,
        put: ViewFn[P, R] | None = None,
        patch: ViewFn[P, R] | None = None,
        delete: ViewFn[P, R] | None = None,
    ) -> Route:
        _pattern = URLPatternStr(url_pattern)
        if _pattern in self.routes:
            raise ImproperlyConfigured(
                f"Cannot add route {_pattern} to router. Route already exists."
            )
        # Create route
        name = f"{self.basename}-{name}" if self.basename is not None else name
        route = Route(name)
        # Register any provided views
        views: dict[HttpMethod, ViewFn[P, R] | None] = {
            HttpMethod.GET: get,
            HttpMethod.POST: post,
            HttpMethod.PUT: put,
            HttpMethod.PATCH: patch,
            HttpMethod.DELETE: delete,
        }
        for method, view in views.items():
            if view is not None:
                route.register({method})(view)
        # Save and return the view
        self.routes[_pattern] = route
        return self.routes[_pattern]

    @property
    def urls(self) -> list[URLPattern]:
        return [
            path(pattern, route.dispatch, name=route.name)
            for pattern, route in self.routes.items()
        ]
